# lavabot2

# to do:
#  -add error handling to Cfg_file.pm
#  -check rfc and modify accordingly
#  -replace 'lavabot' with IRC::get_info(1), add channel flexibility
#  -add valid_handle checks where appropriate
#  -use grep and map more!

$_version = "0.7";
$_xchatversion = IRC::register("LavaBot",$_version,"ircsub_unload","");
IRC::print("---=-=- LavaBot v$_version -=-=---\n");

my (@oplist, $_adminmask, $_adminpwd, $configfile, $_myhandle);

use Msg_parse;
use Cfg_file;

BEGIN {   ### detect modules
	if ($_compiled) { return; }
	$_compiled = "true";     #xchat runs this sub twice otherwise
	$/ = "\n";  # why doesn't xchat/perl set this by default?  i don't know
	my $temp = `ls \*.pm`;
	my @pmfiles = split(".pm\n",$temp);
	my $file;
	foreach $file (@pmfiles) {
		open(MODULE, $file . ".pm") or die "Couldn't open file $file!";
		my $firstline = <MODULE>;
		close(MODULE);
		if ($firstline =~ /:LavaBot:/i) {
			push @modfiles, $file;
			eval "use $file ()";
		}
	}
}

&load_cfg;
IRC::print("mods: ", join(",",@modfiles),"\n");
IRC::add_message_handler('PRIVMSG', ircsub_privmsg);
IRC::add_message_handler('311', ircsub_whois);
IRC::add_message_handler('JOIN', ircsub_join);  ## upon succes of JOIN find handle
foreach $temp (@modfiles) {
	eval "\&$temp\:\:init(\$_channel)";
}

sub load_cfg {
	$configfile = Cfg_file->cfg_newfile("lavabot.cfg");
	local %masterblock = $configfile->cfg_get_block("lavabot","master");
	@oplist = split(";",$masterblock{'ops'});
	$_adminmask = $masterblock{'adminmask'};
	$_adminpwd = $masterblock{'adminpwd'};
	$_channel = $masterblock{'channel'};
	$_myhandle = $masterblock{'my_handle'};
	IRC::print("Lavabot cfg loaded.\n");
}

sub write_cfg {
	local %masterblock = ('ops', join(";",@oplist), 'adminmask', $_adminmask, 'adminpwd', $_adminpwd, 'channel', $_channel, 'my_handle', $_myhandle);
	$configfile->cfg_set_block(\%masterblock, ("lavabot","master"));
	$configfile->cfg_write;
}

sub opped {
### pass in: explicit handle
### returns: 1 if that person is an operator, 0 if not
    my $handle = shift;
    my $opmask;
    foreach $opmask (@oplist) {
        if (&mask_match($opmask, $handle)) {
            return 1;
        }
    }
	# last chance check: if they're admin, we'll consider them an op
    return &is_admin($handle);
}

sub is_admin {
### pass in: explicit handle
### returns: 1 if that person is the admin, 0 if not
	return &mask_match($_adminmask, (shift));
}

sub my_handle {
### self-explanatory accessor
	return $_myhandle;
}

sub ircsub_unload {
	my $i;
	for ($i = $#modfiles; $i >= 0; $i--) {
		my $mod = $modfiles[$i];
		if (!&unload_mod($mod)) {
			IRC::print("Stop: unloaded $mod\n");
		} else {
			IRC::print("WARNING: unable to unload $mod!\n");
		}
	}
	IRC::print("LavaBot: writing to file... ");
	&write_cfg;
	IRC::print("Done unloading.\n");
	return;
}

sub ircsub_join {
	### will (hopefully) trigger the ircsub_whois subroutine  momentarily
	my $joinname = &get_name(shift);
	my $name = IRC::get_info(1);
	if ($joinname eq $name) {   ### it's me!
		IRC::command("/whois " . $name);
	}
}

sub ircsub_whois {
### called upon every '311' message (whois query). will update the return
###  val of 'my_handle'
	my $line = shift;
	my @line = split(" ",$line);
	shift @line; shift @line; shift @line;
	if ($line[0] eq IRC::get_info(1)) {   ### this is a 'whois lavabot'
		$_myhandle = $line[0] . "!" . $line[1] . "\@" . $line[2];
		IRC::print("***** _myhandle updated to: " . $_myhandle . "\n");
	}
}

sub ircsub_privmsg {
	my $line = shift;
	$line =~ s/\s+/ /g;  # collapse excess whitespace
	### check target
	if (&get_target($line) eq $_channel) {
		foreach $temp (@modfiles) {  # pass off to mods
			eval "\&$temp\:\:chanmsg(\$line)";
		}
	} else {
		my $tempname = IRC::get_info(1);
		if (&get_target($line) =~ /^$tempname$/ig) {
			&line_lavabot($line);
		} else {
			IRC::print("UNKNOWN MESSAGE FROM", &get_target($line),"\n");
		}
	}
	return;
}

sub load_mod {
### pass in: name of a mod
### returns: 1 if not found
	my $mod = shift;
	my $i;
	for ($i = 0; $i <= $#unmodfiles; $i++) {
		my $name = $unmodfiles[$i];
		if ($name =~ /$mod/i) {
			push @modfiles, $name;
			splice @unmodfiles, $i, 1;
			eval "\&$name\:\:init(\$_channel)";
			return;
		}
	}
	return 1;
}

sub unload_mod {
### pass in: name of a mod
### returns: 1 if not found
	my $mod = shift;
	my $loc = &locate_mod($mod);
	if ($loc >= 0) {
		my $name = $modfiles[$loc];
		eval "\&$name\:\:stop";
		push @unmodfiles, $name;
		splice @modfiles, $loc, 1;
		### now make sure no one is operating that mod
		my @keys = keys %operators;
		my @values = values %operators;
		my $i;
		for ($i = 0; $i <= $#keys; $i++) {
			if ($values[$i] eq $name) {
				$operators{$keys[$i]} = 0;
			}
		}
		return;
	}
	return 1;
}

sub locate_mod {
### pass in: name of a mod (eg "censor")
### returns: index (in @modfiles) of that mod, -1 if not found
	my $mod = shift;
	for ($i = 0; $i <= $#modfiles; $i++) {
		if ($mod =~ /$modfiles[$i]/i) {
			return $i;
		}
	}
	return -1;
}

sub line_lavabot {
### pass in: a line whose target is lavabot
	my $line = shift;
	my $handle = &get_handle($line);
	my $name = &get_name_from_handle($handle);
	my $shorthandle = get_login($line) . get_domain($line);
	unless (exists $operators{$shorthandle}) {
		# add them to the list with no default mod
		%operators = (%operators, $shorthandle => 0);
	}
	my $body = &get_body($line);
	if (substr($body,0,1) eq ":") {
		# a command with a preceding : is a cmd for the lavabot
		&cmd_lavabot($handle, substr($body,1));
	} elsif ($operators{$shorthandle} eq 0) {
		IRC::command("/msg $name You have not selected a module.  Type ':help' for global commands.\n");
	} else {
		# pass their command to their mod
		eval "\&$operators{$shorthandle}\:\:privmsg(\$handle, \$body)";

	}
	
}

sub cmd_lavabot {
### pass in: explicit handle, body of message (as scalar)
	my $handle = shift;
	my $body = shift;
	my @body = split(" ",$body);
	my $cmd = shift @body;
	my $name = &get_name_from_handle($handle);
	if ($cmd eq "help") {
		IRC::command("/msg $name Global commands are preceded by a colon (:).  Public commands are...");
		IRC::command("/msg $name :help  - this message");
		IRC::command("/msg $name :ver   - display version info");
		IRC::command("/msg $name :mod   - change module");
		IRC::command("/msg $name :listops - list operators");
		if (&is_admin($handle)) {
			IRC::command("/msg $name Admin cmds> :load, :unload, :addop, :delop, :commit, :dump, :admin");
		}
		# IRC::command("/msg $name :load, :unload - load modules");
	} elsif ($cmd eq "ver") {
		IRC::command("/msg $name X-Chat version $_xchatversion. LavaBot version $_version. Loaded modules: " . join(", ",@modfiles) . ".");
	} elsif ($cmd eq "mod") {
		&cmd_mod($handle, @body);
	} elsif ($cmd eq "admin") {
		&cmd_admin($handle, @body);
	} elsif ($cmd eq "load") {
		&cmd_load($handle, @body);
	} elsif ($cmd eq "unload") {
		&cmd_unload($handle, @body);
	} elsif ($cmd eq "addop") {
		&cmd_addop($handle, @body);
	} elsif ($cmd eq "delop") {
		&cmd_delop($handle, @body);
	} elsif ($cmd eq "listops") {
		IRC::command("/msg $name Ops = " . (join(", ",@oplist)));
	} elsif (($cmd eq "commit") and (&is_admin($handle))) {
		&write_cfg;
		IRC::command("/msg $name LavaBot's changes written to disk.");
	} elsif ($cmd eq "dump") {
		IRC::command("/msg $name unmod = " . (join(";",@unmodfiles)) . ", mod = " . (join(";",@modfiles)));

	} else {
		IRC::command("/msg $name Command :$cmd not found. Try :help");
	}
}

sub cmd_admin {
### pass in: handle, mask, passwd
### this will set a new admin mask, provided the correct passwd is supplied
	my $handle = shift;
	my $mask = shift;
	my $passwd = shift;
	my $name = &get_name_from_handle($handle);
	unless ($passwd eq $_adminpwd) {
		IRC::command("/msg $name Bad admin password.  Syntax :admin <mask> <adminpwd>");
		return;
	}
	$_adminmask = $mask;
	IRC::command("/msg $name Adminmask changed to $mask.  Proceed with caution.");
}

sub cmd_delop {
### pass in: handle, mask
        my $handle = shift;
        my $mask = shift;
	my $name = &get_name_from_handle($handle);
	unless (&is_admin($handle)) {
		IRC::command("/msg $name You're not an admin.");
		return;
	}
        for ($i = $#oplist; $i >= 0; $i--) {
                if ($mask eq $oplist[$i]) {
                        splice @oplist, $i, 1;
                        IRC::command("/msg $name $mask deleted from oplist.");
                        return;
                }
        }
        IRC::command("/msg $name Err: $mask not found in oplist.");
}

sub cmd_addop {
### pass in: handle, mask
        my $handle = shift;
        my $mask = shift;
	my $name = &get_name_from_handle($handle);
	unless (&is_admin($handle)) {
		IRC::command("/msg $name You're not an admin.");
		return;
	}
        unless (&valid_handle($mask)) {
                IRC::command("/msg $name Err: bad handle");
                return;
        }
        my $op;
        foreach $op (@oplist) {
                if ($op eq $mask) {
                        IRC::command("/msg $name Err: $mask already on list.");
                        return;
                }
        }
        push @oplist, $mask;
        IRC::command("/msg $name $mask added to oplist.");
        return;
}

sub cmd_load {
### pass in: handle, body
	my $handle = shift;
	my $body = shift;
	my $name = &get_name_from_handle($handle);
	unless (&is_admin($handle)) {
		IRC::command("/msg $name You're not an admin.");
		return;
	}
	if (&load_mod($body)) {
		IRC::command("/msg $name Mod not found, or already loaded.");
	} else {
		IRC::command("/msg $name Mod loaded.");
	}

}

sub cmd_unload {
### pass in: handle, body
	my $handle = shift;
	my $body = shift;
	my $name = &get_name_from_handle($handle);
        unless (&is_admin($handle)) {
                IRC::command("/msg $name You're not an admin.");
                return;
        }
	if (&unload_mod($body)) {
		IRC::command("/msg $name Mod not found, or not loaded.");
	} else {
		IRC::command("/msg $name Mod unloaded.");
	}

}

sub cmd_mod {
### pass in: handle, body
	my $handle = shift;
	my $mod = shift;
	my $name = &get_name_from_handle($handle);
	my $shorthandle = &get_login_from_handle($handle) . &get_domain_from_handle($handle);
	if ($mod) {
		my $index = &locate_mod($mod);
		if ($index < 0) {
			IRC::command("/msg $name Mod $mod not found.");
		} else {
			$operators{$shorthandle} = $modfiles[$index];
			IRC::command("/msg $name You are now operating $operators{$shorthandle}.");
		}
	} else {
		$mod = $operators{$shorthandle};
		if ($mod) {
			IRC::command("/msg $name Current module: $operators{$shorthandle}.");
		} else {
			IRC::command("/msg $name No module currently selected.");
		}
		IRC::command("/msg $name Available modules: " . join(", ",@modfiles) . ".");
	}
	return;
}


