# Censor.pm -- :LavaBot:Censor:

package		Censor;

use Msg_parse;
use Cfg_file;

my $_cfgfilename = "censor.cfg";
my $_version = "0.5";
my $_timer = 1000;

my ($_channel, %c_master, @c_data, @arq_all1, @arq_all2, $spam_last, $spam_last_count, $_alive); 

sub init {
### inits global variables and loads cfg
	$_channel = shift;
	IRC::print("Censor: init... ");
	%c_master = ();
	@c_data = ();
	&c_load_cfg;
	IRC::print("read cfg... ");
	@arq_all1 = ();
	@arq_all2 = ();
	$spam_last = 0;
	$spam_last_count = 0;
	IRC::add_timeout_handler($_timer, Censor::c_timer);
	$_alive = "true";
	IRC::print("done.\n");
}

sub c_gen_actn_template {
	my $ref = 	{ devoice => "true",
			  deop => "true",
			  kick => 0,
			  ban => 0,
			};
	return $ref;
}

sub c_gen_target_template {
	my $ref =	{ mask => 0,
			  actn => &c_gen_actn_template,
			  limit => 1,
			  ar_msg_q => [],
			  ar_msg_prio => [],
			};
	return $ref;
}

sub c_load_cfg {
### takes care of all the OO nonsense that Cfg_file requires
	$c_cfgfile = Cfg_file->cfg_newfile($_cfgfilename);
	my @block = $c_cfgfile->cfg_get_block("censor","master");
	%c_master = @block;
	my $i;
	for ($i = 1; $i <= $c_master{targets}; $i++) {
		my $targnum = "target" . $i;
		my %target = %{ &c_gen_target_template};
		@block = $c_cfgfile->cfg_get_block("censor",$targnum,"main");
		  # this overwrites the default settings (from
		  #  c_gen_target_template) w/ those in the cfg file
		%target = (%target, @block);
		@block = $c_cfgfile->cfg_get_block("censor",$targnum,"actn");
		%{ $target{actn} } = (%{ $target{actn} }, @block);
		my $temp = &c_add_target(%target);
		if ($temp) {
			IRC::print("C_LOAD_CFG $temp\n");
		}
		# add err-check w/ Cfg_file
	}
	
}

sub c_write_cfg {
	my $counter = 1;
	my %temphash = ();
	# clear the file
	$c_cfgfile->cfg_set_block(\%temphash, ("censor"));
	$c_master{targets} = $#c_data + 1;
	$c_cfgfile->cfg_set_block(\%c_master, ("censor","master"));

	foreach $target (@c_data) {
		my %temptarg = %$target;
		my %tempactn = %{ $temptarg{actn} };
		delete $temptarg{actn};
		delete $temptarg{ar_msg_q};
		delete $temptarg{ar_msg_prio};
		$c_cfgfile->cfg_set_block(\%temptarg, ("censor","target$counter","main"));
		$c_cfgfile->cfg_set_block(\%tempactn, ("censor","target$counter","actn"));
	$counter++;
	}
	$c_cfgfile->cfg_write;

}

sub c_add_target {
### pass in: a target record
### returns: 0 if successful, a status message otherwise

	my %target = @_;
	my $mask = $target{mask};
	my $insloc = -1;
	my $i;
	for ($i = 0; $i <= $#c_data; $i++) {
		if (&mask_match($mask, $c_data[$i]{mask})) {
			$insloc = $i;
			last;
		}
	}
	if ($insloc < 0) {
		push @c_data, \%target;
		return;
	} elsif ($mask eq $c_data[$insloc]{mask}) {
		return "Err: duplicate mask at $insloc";
	} else {
		splice(@c_data, $insloc, 0, (\%target));
		return;
	}
}

sub c_find_target {
### pass in: explicit handle 
### returns: index in c_data[?]{mask} of the highest indexed matching target,
### -1 if no match 
	my $handle = shift;
	my $i = 0;
	for ($i = $#c_data; $i >= 0; $i--) {
		# IRC::print("c_find_target: handle = $handle, mask = $c_data[$i]{mask}\n");
		if (mask_match($c_data[$i]{mask}, $handle)) {
			return $i;
		} 	
	}
	return -1;
}

sub c_take_action {
### pass in: handle, taunt string
	my $handle = shift;
	my $taunt = shift;
	my $index = &c_find_target($handle);
	if ($index < 0) {
		return;  ### this person isn't being watched!
	}
	my $mask = "*!*@" . &get_domain_from_handle($handle);
	my $name = &get_name_from_handle($handle);
	if ($c_data[$index]{actn}{devoice}) {
		IRC::command("/mode $_channel -v $name");
	}
	if ($c_data[$index]{actn}{deop}) {
		IRC::command("/mode $_channel -o $name");
	}
	if ($c_data[$index]{actn}{ban}) {
		IRC::command("/mode $_channel +b $mask");
	}
	IRC::command("/msg $_channel $taunt --LavaBot");
	if ($c_data[$index]{actn}{kick}) {
		IRC::command("/kick $name");
	} elsif ($c_data[$index]{actn}{devoice} or $c_data[$index]{actn}{deop}) {
		IRC::command("/mode $_channel +m");
	}
	return;
}

sub chanmsg {
### pass in: a line for processing (straight from lavabot)
	my $line = shift;
	my $last_line = &ar_get_last;
	my $handle = &get_handle($line);
	my $body = &get_body($line);
	if ($body eq &get_body($spam_last)) {
		if ($handle eq &get_handle($spam_last)) {
			$spam_last_count++;
			my $index = &c_find_target($handle);
			if ($index >= 0) {
				if ($spam_last_count >= $c_data[$index]{limit}) {
					&c_take_action($handle, "Shut up you stupid spamming 12 year old!!");
				}
			}
		}
	} else {
		$spam_last_count = 0;
	}
	if ($last_line) {
		if ($handle ne &get_handle($last_line)) {
			if (length $body > 7) {
				&ar_procline($handle, $body);
			}
		}
	}
	$spam_last = $line;
	unshift @arq_all1, $line;
}

sub stop {
### called by lavabot to kill mod
	$_alive = 0;
}

sub ar_find_msg {
### pass in: index of target, msg
### returns: index of msg if already in c_data[]{ar_msg_q}, -1 otherwise
	my ($targetnum, $msg) = @_;
	my $i = 0;
	for ($i = 0; $i <= $#{ $c_data[$targetnum]{ar_msg_q} }; $i++ ) {
		if ($c_data[$targetnum]{ar_msg_q}[$i] eq $msg) {
			return $i;
		}
	} 
	return -1; 
}

sub ar_procline {
### pass in: explicit handle, msg 

	my ($handle, $msg) = @_;
	my $targetnum = &c_find_target($handle);
	if ($targetnum == -1) {
		return -1; #this handle isn't being watched
	}
	my $msgindex = &ar_find_msg($targetnum, $msg);
	if ($msgindex < 0) {
		push @{ $c_data[$targetnum]{ar_msg_q} }, $msg;
		push @{ $c_data[$targetnum]{ar_msg_prio} }, 0;
		# expire messages if more than 20 in q
		if ($#{ $c_data[$targetnum]{ar_msg_q} } > 19) {
			shift @{ $c_data[$targetnum]{ar_msg_q} };
			shift @{ $c_data[$targetnum]{ar_msg_prio} };
		}
	} else {
		my $prio = ++$c_data[$targetnum]{ar_msg_prio}[$msgindex];
		if ($prio >= $c_data[$targetnum]{limit}) {
			&c_take_action($handle, "No Annoying Scripts!");
		}
	}
}

sub ar_get_last {
### returns: last line between q_all1 and 2
	if ($#arq_all1 == -1) {
		if ($#arq_all2 == -1) {
			return 0;
		}
		return $arq_all2[0];
	}
	return $arq_all1[0];
}

sub ar_q_count {
### pass in: index, q number
### returns: number of msgs in that q
	my $index = shift;
	my $q = shift;
	my $count = 0;
	my $temp;
	foreach $temp (@{ $c_data[$index]{ar_msg_prio} }) {
		if ($temp == $q) {
			$count++;
		}
	}
	return $count;

}

sub c_timer {
	unless ($_alive) {
		return;
	}
	@arq_all2 = @arq_all1;
	@arq_all1 = ();
	IRC::add_timeout_handler($_timer, Censor::c_timer);
	return;
}

sub c_validindex {
### pass in: an index
### returns: the index if valid, -1 if not
	my $index = shift;
	unless (defined $index) {
		return -1;
	}
	if ($index =~ /[^\d]/) {
		return -1;
	}
	if (($index < 0) or ($index > $#c_data)) {
		return -1;
	}
	return $index;
}

sub c_opped {
### pass in: explicit handle
### returns: 1 if that person is an operator, 0 if not. Also sends an error.
	my $handle = shift;
	if (&main::opped($handle)) {
		return 1;	
	} else {
		my $name = &get_name_from_handle($handle);
		IRC::command("/msg $name You're not an operator.");
		return 0;
	}
}

sub privmsg {
### pass in: handle, body
### checks for ops before issuing dangerous cmds
	my $handle = shift;
	my $body = shift;
	my $name = &get_name_from_handle($handle);
	my @body = split(" ",$body);
	my $cmd = shift @body;
	if ($cmd eq "help") {
		IRC::command("/msg $name Censor mod $_version.  Public commands are: listtargets");
		IRC::command("/msg $name Operator commands are: addtarget <mask> [high|med|low|none], deltarget <mask|index>, setlimit <index> <limit>, setactn <index> <VOKB flags>, commit");
		IRC::command("/msg $name Advanced commands are: listlowq <index>, listhighq <index>, wipeq <index> <prio>, settimer <timer>");
	} elsif ($cmd eq "listtargets") {
		&c_cmd_listtargets($name);
	} elsif ($cmd eq "addtarget") {
		&c_cmd_addtarget($name, @body) if &c_opped($handle);
	} elsif ($cmd eq "deltarget") {
		&c_cmd_deltarget($name, @body) if &c_opped($handle);
	} elsif ($cmd eq "commit") {
		if (&c_opped($handle)) {
			&c_write_cfg;
			IRC::command("/msg $name Changes commited to disk.");
		}
	} elsif ($cmd eq "setlimit") {
		&c_cmd_setlimit($name, @body) if &c_opped($handle);
	} elsif ($cmd eq "setactn") {
		&c_cmd_setactn($name, @body) if &c_opped($handle);
	} elsif ($cmd eq "settimer") {
		&c_cmd_settimer($name, @body) if &c_opped($handle);
	} elsif ($cmd eq "listlowq") {
		&c_cmd_listlowq($name, @body) if &c_opped($handle);
	} elsif ($cmd eq "listhighq") {
		&c_cmd_listhighq($name, @body) if &c_opped($handle);
	} elsif ($cmd eq "wipeq") {
		&c_cmd_wipeq($name, @body) if &c_opped($handle);
	} else {
		IRC::command("/msg $name Syntax Err: Command not found.  Try help");
	}

}

sub c_cmd_wipeq {
### pass in: name, index, q num
	my $name = shift;
	my $index = &c_validindex(shift);
	my $q = shift;
	if (!defined $q) {
		$q = 0;
	}
	if ($index < 0) {
		IRC::command("/msg $name Err: bad index");
		return;
	}
	my $count = 0;
	my $i;
	for ($i = $#{ $c_data[$index]{ar_msg_q}}; $i >= 0; $i--) {
		if ($c_data[$index]{ar_msg_prio}[$i] == $q) {
			splice @{ $c_data[$index]{ar_msg_q}}, $i, 1;
			splice @{ $c_data[$index]{ar_msg_prio}}, $i, 1;
			$count++;
		}
	}
	IRC::command("/msg $name Wiped $count msgs from q $q for target $index");
	return;
}

sub c_cmd_listlowq {
### pass in: name, index
	my $name = shift;
	my $index = &c_validindex(shift);
	if ($index < 0) {
		IRC::command("/msg $name Err: bad index");
		return;
	}
	my $count = &ar_q_count($index, 0);
	IRC::command("/msg $name count is $count for this index.");
	if ($count > 9) {
		return;
	}
	my $i;
	for ($i = 0; $i <= $#{ $c_data[$index]{ar_msg_q}}; $i++) {
		if ($c_data[$index]{ar_msg_prio}[$i] == 0) {
			IRC::command("/msg $name $c_data[$index]{ar_msg_q}[$i]");
		}
	}
	IRC::command("/msg $name -- End of low Q listing for #$index --");
	return;
}

sub c_cmd_listhighq {
### pass in: name, index
	my $name = shift;
	my $index = &c_validindex(shift);
	if ($index < 0) {
		IRC::command("/msg $name Err: bad index");
		return;
	}
	my $i;
	for ($i = 0; $i <= $#{ $c_data[$index]{ar_msg_q}}; $i++) {
		my $prio = $c_data[$index]{ar_msg_prio}[$i];
		if ($prio > 0) {
			IRC::command("/msg $name $prio : $c_data[$index]{ar_msg_q}[$i]");
		}
	}
	IRC::command("/msg $name -- End of high Q listing for #$index --");
	return;
}

sub c_cmd_settimer {
### pass in: name, timerval
	my $name = shift;
	my $timer = shift;
	if (($timer < 500) or ($timer > 5000)) {
		IRC::command("/msg $name Err: timer val must be 500 - 5000.");
		return;
	}
	$_timer = $timer;
	return;
}

sub c_cmd_setlimit {
### pass in: name, target index, limit
	my $name = shift;
	my $index = &c_validindex(shift);
	my $limit = shift;
	unless (defined $limit) {
		IRC::command("/msg $name Err: Syntax> setlimit <index> <limit>");
		return;
	}
	if ($index < 0) {
		IRC::command("/msg $name Err: bad index");
		return;
	} elsif  ($limit < 1) {
		IRC::command("/msg $name Err: bad limit value.  Limit must be > 0");
	}
	$c_data[$index]{limit} = $limit;
	IRC::command("/msg $name Limit of #$index set to $limit.");
	return;
}

sub c_cmd_setactn {
### pass in: name, index, flags (eg "VOKB")
	my $name = shift;
	my $index = &c_validindex(shift);
	my $flags = shift;
	if ((!$flags) or ($flags =~ /[^(v|o|k|b)]/i)) {
		IRC::command("/msg $name Syntax Error: setactn <index> <VOKB flags>");
		return;
	}
	if ($index < 0) {
		IRC::command("/msg $name Err: bad index");
		return;
	}

	$c_data[$index]{actn}{devoice} = ($flags =~ /v/i) ? "true" : 0;
	$c_data[$index]{actn}{deop} = ($flags =~ /o/i) ? "true" : 0;
	$c_data[$index]{actn}{kick} = ($flags =~ /k/i) ? "true" : 0;
	$c_data[$index]{actn}{ban} = ($flags =~ /b/i) ? "true" : 0;
	IRC::command("/msg $name Index ${index}'s flags set to $flags");
	return;
}

sub c_cmd_deltarget {
### pass in: name, and index of target or an identical mask
	my $name = shift;
	my $arg = shift;
	if (&c_validindex($arg) >= 0) {
		IRC::command("/msg $name Target $c_data[$arg]{mask} deleted.");
		splice @c_data, $arg, 1;
		return;
	} else {
		for ($i = $#c_data; $i >= 0; $i--) {
			if ($arg eq $c_data[$i]{mask}) {
				splice @c_data, $i, 1;
				IRC::command("/msg $name Target $arg deleted.");
				return;
			}
		}
	}
	IRC::command("/msg $name Err: target not found.");
}

sub c_cmd_addtarget {
### pass in: name, target mask, and (optionally) high/med/low/none
	my $name = shift;
	my $mask = shift;
	my $class = shift;
	my %targ = %{ &c_gen_target_template};
	$targ{mask} = $mask;
	## move these cases into c_gen_target_template?
	if ($class eq "high") {
		$targ{limit} = 1;
		$targ{actn}{devoice} = "true";
		$targ{actn}{deop} = "true";
		$targ{actn}{kick} = "true";
		$targ{actn}{ban} = "true";
	} elsif ($class eq "med") {
		$targ{limit} = 2;
		$targ{actn}{devoice} = "true";
		$targ{actn}{deop} = "true";
		$targ{actn}{kick} = "true";
		$targ{actn}{ban} = 0;
	} elsif ($class eq "low") {
		$targ{limit} = 4;
		$targ{actn}{devoice} = "true";
		$targ{actn}{deop} = "true";
		$targ{actn}{kick} = 0;
		$targ{actn}{ban} = 0;
	} elsif ($class eq "none") {
		$targ{limit} = 99;
		$targ{actn}{devoice} = 0;
		$targ{actn}{deop} = 0;
		$targ{actn}{kick} = 0;
		$targ{actn}{ban} = 0;
	}
	my $result = &c_add_target(%targ);
	if ($result) {
		IRC::command("/msg $name AddTarget $result");
	} else {
		IRC::command("/msg $name Target successfully added ($mask)");
	}
}

sub c_cmd_listtargets {
### pass in: name
	my $name = shift;
	IRC::command("/msg $name Index Lim  Actn  Mask");
	for ($i = 0; $i <= $#c_data; $i++) {
		my @targstr = ();
		push @targstr, "$i.";
		push @targstr, $c_data[$i]{limit};
		my @actnstr = ();
		if ($c_data[$i]{actn}{devoice}) {
			push @actnstr, "V";
		}
		if ($c_data[$i]{actn}{deop}) {
			push @actnstr, "O";
		}	
		if ($c_data[$i]{actn}{kick}) {
			push @actnstr, "K";
		}
		if ($c_data[$i]{actn}{ban}) {
			push @actnstr, "B";
		}
		my $actn = join("",@actnstr);
		push @targstr, $actn;
		push @targstr, $c_data[$i]{mask};
		IRC::command("/msg $name " . join("    ",@targstr));
	}
}

1;  #fucking compatibility
