# Cfg_file.pm
package		Cfg_file;

my $_treechar = "]";

### This file takes full advantage of perl's fucked up OO abilities.
### cfg_newfile (the constructor) returns a ref to an array of 3 refs
###  1.) a ref to the name of the file (eg "myconfig.cfg")
###  2.) a ref to an array called @cfg_file that holds a 'clean' copy
###   of the entire cfg file
###  3.) a hash, cfg_file_ref, for which each key is join($_treechar,@branch)
###    and each value is the offset (in @cfg_file) of that branch


sub cfg_newfile {
### pass in: filename
### this is the constructor
	my $classname = shift;
	my $filename = shift;
	open(CFG, $filename) or die "Couldn't open $filename!";
	my @cfg_file = ();
	my %cfg_file_ref = ();
	while (<CFG>) {
		s/\s+/ /g;
		s/(^\s)|(\s$)//g;
		$temp = $_;
		
		if (length > 0) {
			push @cfg_file, $temp;
		}
	}
	close(CFG) or die "Couldn't close the goddamn file!";
	my $self = [\$filename, \@cfg_file, \%cfg_file_ref];
	bless $self;
	$self->_cfg_verify;
	return $self;
}

sub cfg_get_block {
### pass in: block tree (eg "Censor","target1","actn")
### returns: an array/hash of the keys and values in that block
	my $self = shift;
	my @p_tree = @_;

	my $cfg_file = $self->[1];
	my $cfg_file_ref = $self->[2];

	my $path = join($_treechar, @p_tree);
	unless (exists $cfg_file_ref->{$path}) {
		IRC::print("ERR: get_block key not found: $path\n");
		# set some cfg_error?
		return ();
	}
	my @block = ();
	my $index = $cfg_file_ref->{$path};
	while (substr($cfg_file->[++$index],0,1) ne '[') {
		my $line = $cfg_file->[$index];
		my @entries = split " = ", $line;
		if ($entries[1] eq "false") {
			$entries[1] = 0;
		}
		if ($#entries != 1) {
			IRC::print("ERR: bad entry: $line\n");
			next;
		}
		push @block, @entries;
	}
	return @block;
}

sub cfg_set_block {
### pass in: reference to a hash, block tree (eg "Censor","target1")
	my $self = shift;
	my $blockref = shift;
	my @exist_tree = @_;

	my $cfg_file = $self->[1];
	my $cfg_file_ref = $self->[2];

	my @build_tree = ();
	while (!exists $cfg_file_ref->{join($_treechar, @exist_tree)} && @exist_tree) {
		my $temp = join($_treechar, @exist_tree);
		push @build_tree, pop @exist_tree;
	}
	my $offset = 0;
	if (@exist_tree) {
		$offset = $cfg_file_ref->{join($_treechar, @exist_tree)} + 1;
	}
	if (@build_tree) {
		foreach $branch (reverse @build_tree) {
			splice @$cfg_file, $offset++, 0, ("[" . $branch . "]", "[/" . $branch . "]");
		}
	} else {
		my $endtarg = "[/$exist_tree[$#exist_tree]]";
		while ($cfg_file->[$offset] ne $endtarg) {
			splice @$cfg_file, $offset, 1;
		}

	}
	my $key = 0;
	my $value = 0;
	while (($key, $value) = each %$blockref) {
		splice @$cfg_file, $offset++, 0, $key . " = " . $value;
	}
	$self->_cfg_verify;

}

sub cfg_write {
	my $self = shift;
	my $filename = $self->[0];
	my $cfg_file = $self->[1];

	open(CFG, ">$$filename") or die "Couldn't write to the file!";
	my $last = "";
	foreach $temp (@$cfg_file) {
		if ($last =~ /^\[.*\]$/ && $temp =~ /^\[.*\]$/) {
			print CFG "\n";
		}
		print CFG $temp . "\n";
		$last = $temp;
	}
	print CFG "\n";
	close(CFG) or IRC::print("Couldn't close the goddamn file!");
}

sub _cfg_verify {
### analyzes @cfg_file and builds %cfg_file_ref
	my $self = shift;
	my @cfg_file = @{ $self->[1]};

	my %cfg_file_ref = ();
	my @p_tree = ();
	my $i;
	for ($i = 0; $i <= $#cfg_file; $i++) {
		if ($cfg_file[$i] =~ /^\[(.*)\]$/) {
			$header = $1;
			if (substr($header,0,1) eq '/') {
				('/' . (pop @p_tree) eq $header) or IRC::print("ERR: bad ender $header\n");
			} else {
				push @p_tree, $header;
				my $key = join $_treechar, @p_tree;
				if (exists $cfg_file_ref{$key}) {
					IRC::print("ERR: dup key $key\n");
				}
				$cfg_file_ref{$key} = $i;
			}
		}
	}
	if ($#p_tree >= 0) {
		IRC::print("ERR: unclean end of tree\n");
		while ($#p_tree >= 0) {
			my $ender = "[/" . (pop @p_tree) . "]";
			push @cfg_file, $ender;
		}
	}
	%{ $self->[2]} = %cfg_file_ref;
}

1;

