import java.net.*;
import java.io.*;
import java.util.*;

// The Log class in this file is a general purpose
// file analyzer for line-formatted files.
// The interfannce LineProcessor specifies a method
// ProcessLine(s) that tells the line reader what
// to do with each line.
// The static method ProcessLines processes any
// text file with any LineProcessor interface.
// 

interface LineProcessor {
  final int LineMax=1000000;
  public void ProcessLine(String s);
}

// the simples LineProcessor is the line counter:

class LineCounter implements LineProcessor {
int linecnt=0;
    public int getLineCount () {
      return linecnt;
    }
    public void ProcessLine(String s) {
      linecnt++;
    }
}

class LineSaver implements LineProcessor {
PrintStream ps;
    public void setPS (String f) {
      try {
        FileOutputStream fos = new FileOutputStream(f);
        ps = new PrintStream(fos); 
      } catch (Exception e) {System.out.println("LINESAVER: "+e);}
    }
    public void ProcessLine(String s) {
        int n = s.indexOf(" t:");
        if (n > 0) {
            String line = s.substring(n+3);
	    ps.println(line);
        }
    }
}

class FastLoader implements LineProcessor {

protected String pattern="";
protected String that="*";
protected String template="";
protected String category="";
protected String filename="";

public void ProcessLine(String s) {
  int n; 
  //if ((n=s.indexOf("<alice>")) >= 0) {System.out.println(n+"ALICE");}
  //if ((n=s.indexOf("</alice>")) >= 0) {System.out.println(n+"/ALICE");}
  if ((n=s.indexOf("<category>")) >= 0) {
     s = "";  // assumping <category> alone on line
     category = s;
  }
  if ((n=s.indexOf("</category>")) >= 0) {
     String stag="<pattern>"; String etag ="</pattern>";  
     if (category.indexOf(stag) >= 0) {
       int start = category.indexOf(stag);
       int end = category.indexOf(etag);
       if (end <= start) System.out.println("Something funny");
       else {
         String mid = category.substring(start+stag.length(), end);
         pattern = mid;
       }
     }
     stag="<that>"; etag ="</that>";  
     if (category.indexOf(stag) >= 0) {
       int start = category.indexOf(stag);
       int end = category.indexOf(etag);
       if (end <= start) System.out.println("Something funny");
       else {
         String mid = category.substring(start+stag.length(), end);
         that = mid;
       }
     }
     stag="<template>"; etag ="</template>";  
     if (category.indexOf(stag) >= 0) {
       int start = category.indexOf(stag);
       int end = category.indexOf(etag);
       if (end <= start) System.out.println("Something funny");
       else {
         String mid = category.substring(start+stag.length(), end);
         template = mid;
       }
     }
     /*
     System.out.println(pattern);
     System.out.println(that);
     System.out.println(template);
     */
     if (Classifier.brain.size() % 1000 == 0) 
        System.out.println(Classifier.brain.size()+" Categories and Loading");
     Classifier.brain.add(pattern, that, template, filename);
  }
  else {
    category += s+"\n";
  }
}

public static void ProcessLines (String fname, FastLoader lp) {
  FileInputStream fis;
  BufferedReader br;  
  lp.filename=fname;
  try { 
    fis = new FileInputStream(fname);
    br = new BufferedReader(new InputStreamReader(fis));
    String line;
    while ((line = br.readLine()) != null) {
       lp.ProcessLine(line);
    }
  } catch (Exception e) {
    System.out.println("FastLoader.ProcessLines "+e);
  }
}  

} // class FastLoader

// Log is the class with "main" and the static "ProcessLines"
// methods:

class Log implements LineProcessor {

public void ProcessLine(String s) {

} 

public static void ProcessLines (String fname, LineProcessor lp) {
  ProcessLines (fname, lp, LineProcessor.LineMax) ;
}

public static void ProcessLines (String fname, LineProcessor lp, int limit) {
  ProcessLines (fname, lp, 0, limit);
}

public static void ProcessLines (String fname, LineProcessor lp, 
				 int start, int end) {
  FileInputStream fis;
  BufferedReader br;  
  int cnt=0;
  System.out.println("PL: "+fname+" "+start+" "+end);
  try { 
    fis = new FileInputStream(fname);
    br = new BufferedReader(new InputStreamReader(fis));
    String line;
    while (cnt < end  && 
	   (line = br.readLine()) != null && line.length() >= 0) {
       if (cnt >= start) lp.ProcessLine(line);
       cnt++;
    }
  } catch (Exception e) {
    System.out.println("fname="+fname+" start="+start+" end="+end);
    System.out.println("PROCESSLINES: "+e);
  }
}  

public static void main (String argv[]) {
convert_dont("dont.txt","/home/httpd/html/dont.html");
   //FastLoader fl = new FastLoader();
    //fl.ProcessLines("B.aiml", fl);
//convert_dont("/home/httpd/html/edit.txt","/home/httpd/html/edit.html");
analyze_refers();
//analyze_access();
//analyze_dialog("../superlog");
}

public static void convert_dont(String dontfile, String htmlfile) {
  LineCounter lc = new LineCounter();
  ProcessLines(dontfile, lc);
  int numlines = lc.getLineCount();
  System.out.println(numlines+" lines in don't read me");
  DontHTMLer dh = new DontHTMLer(htmlfile, "Dont.aiml");
  ProcessLines(dontfile, dh, 0, numlines);    
  StringFile sf = new StringFile();
  sf.appendLine("Dont.aiml","<category>");
  sf.appendLine("Dont.aiml","<pattern>HELP</pattern>");
  sf.appendLine("Dont.aiml","<template><random>"+dh.faq+"</random></template>");
  sf.appendLine("Dont.aiml","</category>");
  sf.appendLine("Dont.aiml","<category>");
  sf.appendLine("Dont.aiml","<pattern>FAQ</pattern>");
  sf.appendLine("Dont.aiml","<template><ul>"+dh.faq+"</ul></template>");
  sf.appendLine("Dont.aiml","</category>");
  sf.appendLine("Dont.aiml","</alice>");
}

// I'm assuming a Linux installation of
// an Apache web server
// 
public static void analyze_refers() {
  FileOutputStream fos;
  PrintStream ps;
  try {
  LineCounter lc = new LineCounter();
  //ProcessLines("referer_log", lc);
  //  ProcessLines("/var/log/httpd/referer_log.1", lc);
  ProcessLines("/var/log/httpd/referer_log", lc);
  int numlines = lc.getLineCount();
  System.out.println(numlines+" lines in referer logs");
  //  ProcessLines("/var/log/httpd/referer_log.1", new RefererLog(), 0, numlines);   
  ProcessLines("/var/log/httpd/referer_log", new RefererLog(), 0, numlines); 
  //ProcessLines("referer_log", new RefererLog(), 0, numlines); 
  fos = new FileOutputStream("/home/httpd/html/mine.html");
  ps = new PrintStream(fos);
  ps.println("<H1>ALICE's Top 10 Referers</H1>");
  ps.println("San Francisco, CA ("+new Date()+") <BR>");
  ps.println("The web sites providing the most traffic to ALICE this week were:<BR>"); 
  ps.println("Rank -- Refering Site<BR>"); 
  for (int i = 0; i < RefererLog.S.size(); i++) {
    int n = RefererLog.S.Count(i);
    String s = (String)RefererLog.S.elementAt(i);
    System.out.println(n+" "+s);
    String t = s;
    if (s.indexOf("bots.internet") > 0) t = "Bots.Internet.Com (formerly BotSpot)";
    if (s.indexOf("birch") > 0) t = "Lehigh.edu (old ALICE site includes all traffic from Yahoo/Web AI Games)";
    if (s.indexOf("amused") > 0) t = "Center for the Easily Amused";
    if (s.indexOf("winmag") > 0) t = "Mike Elgan in Windows Magazine";
    if (s.indexOf("magelang") > 0) t = "MageLang Institute";
    if (s.indexOf("/cimh") > 0) t = "Computers in Mental Health";
    if (s.indexOf("robitron") > 0) t = "Robitron, Robby Garner";
    if (s.indexOf("geeknews") > 0) t = "Geek News";
    if (s.indexOf("arstechnica") > 0) t = "Ars Technica PC News";
    if (s.indexOf("vandaveer") > 0) t = "Online Arcades, Java Games";
    if (s.indexOf("jguru") > 0) t = "jGuru.com";
    if (s.indexOf("antlr") > 0) t = "ANTLR Open Source Translater";
    if (s.indexOf("j/w/h") > 0) t = "Southern Humor and Recipes";
    ps.println((i+1)+". ("+n+") <A HREF="+s+">"+s+"</A> "+t+"<BR>");
  }
  ps.println("<HR>");
  } catch (Exception e) {System.out.println("LOG: "+e);}
}

public static void analyze_dialog(String logfile) {
  FileOutputStream fos;
  PrintStream ps;
  try {
  LineCounter lc = new LineCounter();
  LineSaver ls = new LineSaver(); 
  ls.setPS("human.txt");
  ProcessLines(logfile, lc);
  ProcessLines(logfile, ls);
  int numlines = lc.getLineCount();
  System.out.println(numlines+" lines in dialog log");
  ProcessLines(logfile, new DialogLog(), 0, numlines); 
  fos = new FileOutputStream("access.txt");
  ps = new PrintStream(fos);
  int n = DialogLog.S.size();
  int sum = 0;
  for (int i = n-1; i >= 0; i--) {
    sum = sum + DialogLog.S.Count(i);
    System.out.println(DialogLog.S.Count(i)+" "+DialogLog.S.elementAt(i));
    ps.println(DialogLog.S.Count(i)+" "+DialogLog.S.elementAt(i));
  }
  for (int i = 0; i < n; i++) {
    System.out.println(DialogLog.R.elementAt(i));
    ps.println(DialogLog.R.elementAt(i));
  }
  System.out.println(DialogLog.S.size()+" hosts");
  System.out.println("Mean: "+sum/(n+1)+" Median: "+DialogLog.S.Count(n/2));
  ps.println(DialogLog.S.size()+" hosts");
  ps.println("Mean: "+sum/(n+1)+" Median: "+DialogLog.S.Count(n/2));

  } catch (Exception e) {System.out.println("Exception "+e);}
}

public static void analyze_access() {
  FileOutputStream fos;
  PrintStream ps;
  try {
  LineCounter lc = new LineCounter();
  ProcessLines("/var/log/httpd/access_log", lc);
  int numlines = lc.getLineCount();
  System.out.println(numlines+" lines in access log");
  ProcessLines("/var/log/httpd/access_log", new AccessLog(), 0, numlines); 
  fos = new FileOutputStream("access.txt");
  ps = new PrintStream(fos);
  for (int i = AccessLog.S.size()-1; i >= 0; i--) {
    System.out.println(AccessLog.S.Count(i)+" "+AccessLog.S.elementAt(i));
    ps.println(AccessLog.S.Count(i)+" "+AccessLog.S.elementAt(i));
  }
  } catch (Exception e) {System.out.println("LOG: "+e);}
}
} // class 

class DontHTMLer implements LineProcessor {

FileOutputStream fos;
PrintStream ps; 
    PrintStream aps; // AIML File
int linecount = 0;
String filename;
public DontHTMLer(String file, String afile) {
filename = file;
try {
  fos = new FileOutputStream(filename);
  ps = new PrintStream(fos);
  fos = new FileOutputStream(afile);
  aps = new PrintStream(fos);
  aps.println("<alice>");
} catch (Exception e) {System.out.println("LOG: "+e);}
} // constructor

String pattern = "";
String template="";
public String faq = "";

public void ProcessLine(String s) {
if (s.indexOf(">") >= 0) s = Substituter.replace(">","&gt;",s);
if (s.indexOf("<") >= 0) s = Substituter.replace("<","&lt;",s);
if (linecount ==  0) s = "<html><head><title>Don't Read Me: ALICE and AIML Documentation</title></head><body bgcolor=\"#FFFFFF\"><CENTER><H1><EM><B>"+s+"</B></EM></H1></CENTER>";
else if (s.startsWith("\t")) s = "<CENTER><H2>"+s+"</H2></CENTER>";
else if (s.startsWith("I.")) s = "<H4>"+s+"</H4>";
else if (s.startsWith("II.")) s = "<H4>"+s+"</H4>";
else if (s.startsWith("III.")) s = "<H4>"+s+"</H4>";
else if (s.startsWith("IV.")) s = "<H4>"+s+"</H4>";
else if (s.startsWith("V.")) s = "<H4>"+s+"</H4>";
else if (s.startsWith("VI.")) s = "<H4>"+s+"</H4>";
else if (s.startsWith("VII.")) s = "<H4>"+s+"</H4>";
else if (s.startsWith("VIII.")) s = "<H4>"+s+"</H4>";
else if (s.startsWith("IX.")) s = "<H4>"+s+"</H4>";
else if (s.startsWith("X.")) s = "<H4>"+s+"</H4>";
else if (s.startsWith("Appendix")) s = "<H4>"+s+"</H4>";
else if (s.startsWith("Contents")) s = "<H4>"+s+"</H4>";
else if (s.startsWith("- ")) {
  if (template.length() > 0 && pattern.length() > 0) {
    aps.println("<category>");
    aps.println("<pattern>"+pattern+"</pattern>");
    aps.println("<template>"+template+"</template>");
    aps.println("</category>");
  }
  template = "";
  pattern = s;
  faq = faq + "<li>"+s+"<br></li>"+"\n";
  System.out.println(pattern);
  pattern = Substituter.normalize(pattern);
  System.out.println(pattern);
  s = "<H4>"+s+"</H4>";
}
 else {template += s+"<br>\n"; s = s+"<BR>"; }
//System.out.println(s);
ps.println(s);
linecount++;
}
}

  


class RefererLog implements LineProcessor {
static StringRanker S = new StringRanker();
Properties dns = new Properties();
int linecnt=0;

public void ProcessLine(String s) {
  StringTokenizer st = new StringTokenizer(s);
  int n = st.countTokens(); 
  linecnt++;
  String hname = st.nextToken();
  hname=hname.trim();
  if (hname.indexOf("#") > 0) hname = hname.substring(0,hname.indexOf("#"));
  hname = Substituter.replace("%7E","~", hname);
  hname = Substituter.replace("%7e","~", hname);
  hname = Substituter.replace("http://amused","http://www.amused",hname);
  hname = Substituter.replace("http://botspot","http://bots.internet",hname);
  hname = Substituter.replace("http://www.botspot","http://bots.internet",hname);
  hname = Substituter.replace("MageLang","magelang",hname);
  hname = Substituter.replace("ars-technica","arstechnica",hname);
  hname = Substituter.replace(".com/index.html",".com",hname);
  hname = Substituter.replace(".net/index.shtml",".net",hname);
  hname = Substituter.replace(".com/index.shtml",".com",hname);
  hname = Substituter.replace("www.arstechnica","arstechnica",hname);
  hname = Substituter.replace("www.geeknews","geeknews",hname);
  hname = Substituter.replace("207.153.240.12","geeknews.net",hname);
  hname = Substituter.replace("geeknews.net/","geeknews.net",hname);
  hname = Substituter.replace("www2.islandnet.com","www.amused.com",hname);
  hname = Substituter.replace("home.att.net/~r.g.garner","www.robitron.com",hname);
  hname = Substituter.replace("www.fringeware.com/~robitron","www.robitron.com",hname);
  hname = Substituter.replace("parr-research","magelang",hname);
  hname = Substituter.replace("jGuru","jguru",hname);
  hname = Substituter.replace("zug.com/zug","zug.com",hname);
  hname = Substituter.replace("www.xml.com","xml.com",hname);
  hname = Substituter.replace("xml.com/xml","xml.com",hname);
  if (hname.indexOf("robitron") > 0) hname = "http://www.robitron.com";
  if (hname.indexOf("magelang") > 0) hname = "http://www.magelang.com";
  if (hname.indexOf("amused.com") > 0) hname = "http://www.amused.com";
  if (hname.indexOf("lehigh") > 0) hname = "http://birch.eecs.lehigh.edu/alice";
  if (hname.indexOf("oasis") > 0) hname = "XML/SGML Resource Pages";
  if (hname.indexOf("?") < 0 && 
      hname.startsWith("http") &&
      hname.compareTo("-") != 0 &&
      hname.indexOf("rwallace") < 0 &&
      //hname.indexOf("lehigh") < 0 &&
      hname.indexOf("not.likely") < 0 &&
      hname.indexOf("Read") < 0 &&
      hname.indexOf("Message") < 0 &&
      hname.indexOf("cgi-bin") < 0 &&
      hname.indexOf(".gif") < 0 &&
      hname.indexOf(".asp") < 0 &&
      hname.indexOf("mail") < 0 &&
      hname.indexOf("netcraft") < 0 &&
      hname.indexOf("206") < 0 && 
      hname.indexOf("3355") < 0 && 
      hname.indexOf("jguru") < 0 &&
      hname.indexOf("magelang") < 0 &&
      hname.indexOf("Alicebot") < 0 &&
      hname.indexOf("alicebot") < 0 &&
      hname.indexOf("aliceBot") < 0 &&
      hname.indexOf("AliceBot") < 0) {
    System.out.println(hname);
    S.add(hname);
  }
}
}
 
class IPHandler {
String reverse_ip(String s) {
  StringTokenizer st = new StringTokenizer(s,".");
  int n = st.countTokens(); 
  String outstring = "";
  for (int i = 0; i < n; i++) {
    String w = st.nextToken();
    outstring =  w + ((i==0)? "" : ".") + outstring;
  }
  return outstring;
}
} // class IPHandlder


class DialogLog extends IPHandler implements LineProcessor {
static StringRanker S = new StringRanker();
static StringSorter R = new StringSorter();
public void ProcessLine(String s) {
  int n = s.indexOf(" t:");
  if (n > 0) {
    String ip = s.substring(0, n);
    ip = reverse_ip(ip);
  }
} // process line
} // class

class AccessLog extends IPHandler implements LineProcessor {
Properties dns = new Properties();
static StringRanker S = new StringRanker();

int linecnt=0;


public void ProcessLine(String s) {
  StringTokenizer st = new StringTokenizer(s);
  int n = st.countTokens(); 
  linecnt++;
  String w = st.nextToken();
  S.add(w);
  System.out.println(linecnt);
  if (!dns.containsKey(w)) {
     dns_lookup(w);
  }
}

public void dns_lookup (String hostname) {
  try {
    InetAddress ia = InetAddress.getByName(hostname);
    String hname = ia.getHostName();
    hname = reverse_ip(hname);
    dns.put(hostname, hname);
    System.out.println(hname);
  } catch (Exception e) {System.out.println("LOG: "+e);}
} 
}   









